export enum TokenParams {
    none = 0,
    name = 1 << 1,
    symbol = 1 << 2,
    decimals = 1 << 3,
    total_supply = 1 << 4,
    min_transfer = 1 << 5,
    max_transfer = 1 << 6,
    fee_divisor = 1 << 7,
    all = name | symbol | decimals | total_supply | min_transfer | max_transfer | fee_divisor
}
export type TokenConfig = {
    approval_app: number;
    assets_info: Token[];
}
export type Token = {
    network: string;
    name: string;
    symbol: string;
    type: string;
    asset_id: number | undefined;
    bridge_vault_id: number | undefined;
    decimals: number;
    min_transfer: number | undefined;
    max_transfer: number | undefined;
    fee_divisor: number | undefined;
    total_supply: bigint;
}
export function initToken(options?: Partial<Token>): Token {
    const defaults = {
        network: "",
        name: "",
        symbol: "",
        type: "",
        asset_id: undefined,
        bridge_vault_id: undefined,
        decimals: 0,
        min_transfer: undefined,
        max_transfer: undefined,
        fee_divisor: undefined,
        total_supply: BigInt(0)
    }
    return {
        ...defaults,
        ...options,
    };
}
export class Tokens {

    public defined: boolean = false;
    public Token = initToken();

    public constructor(requiredParams?: TokenParams) {

        try {

            //parse params
            var argv = require('minimist')(process.argv.slice(2));
            let token = this.Token;

            token.name = argv.name;
            token.decimals = argv.d;
            token.symbol = argv.s;
            if (argv.network) token.network = argv.network.toLowerCase();
            token.min_transfer = argv.min;
            token.max_transfer = argv.max;
            token.fee_divisor = argv.fee;

            let totalSupply = argv.t as string;
            token.total_supply = (!totalSupply ? BigInt(0) : BigInt(totalSupply));

            //ensure params are set
            if (requiredParams) {
                if ((requiredParams & TokenParams.name) === TokenParams.name) {
                    if (!token.name && token.name === undefined) throw new Error("Token Name is required");
                }
                if ((requiredParams & TokenParams.symbol) === TokenParams.symbol) {
                    if (!token.symbol && token.symbol === undefined) throw new Error("Asset Symbol is required");
                }
                if ((requiredParams & TokenParams.decimals) === TokenParams.decimals) {
                    if (!token.decimals && token.decimals === undefined) throw new Error("Decimals is required");
                }
                if ((requiredParams & TokenParams.total_supply) === TokenParams.total_supply) {
                    if (!token.total_supply && token.total_supply === undefined) throw new Error("Total Supply is required");
                }

                if ((requiredParams & TokenParams.min_transfer) === TokenParams.min_transfer) {
                    if (!token.min_transfer) {
                        throw new Error('Minimum Transfer not specified');
                    } else if (token.min_transfer <= 0) {
                        throw new Error('Minimum Transfer must be greater than 0');
                    }
                }

                if ((requiredParams & TokenParams.max_transfer) === TokenParams.max_transfer) {
                    if (!token.max_transfer) {
                        throw new Error('Maximum Transfer not specified');
                    } else if (token.max_transfer <= 0) {
                        throw new Error('Maximum Transfer must be greater than 0');
                    } else if (!token.min_transfer) {
                        throw new Error('Minimum Transfer not specified');
                    } else if (token.max_transfer <= token.min_transfer) {
                        throw new Error('Maximum Transfer must be greater than Minimum Transfer');
                    }
                }

                if ((requiredParams & TokenParams.fee_divisor) === TokenParams.fee_divisor) {
                    if (!token.fee_divisor) {
                        throw new Error('Fee Divisor not specified');
                    } else if (token.fee_divisor < 1) {
                        throw new Error('Fee Divisor must be greater than 1');
                    }
                }
            }

            //Re-set token
            this.Token = token;

            //Passed All Checks:
            this.defined = true;

        } catch (error) {
            console.log(error);
        }

    }
    public static getConfig(rootPath: string, mode: string): TokenConfig {
        let path = rootPath + `/configs/${mode}.tokenConfig.json`;
        let tokenConfig = require(path) as TokenConfig;
        return tokenConfig;
    }
    public static getToken(config: TokenConfig, network: string, symbol: string): Token | undefined {
        console.log(`matching: ${symbol.toLowerCase()} ${network.toLowerCase()}`);
        
        let tokens = config.assets_info;
        for (let i = 0; i < tokens.length; i++) {
            let token = tokens[i];          
            if (token.network.toLowerCase() === network.toLowerCase() &&
                token.symbol.toLowerCase() === symbol.toLowerCase()) return token;
        }
        return undefined;
    }
    public static saveConfig(rootPath: string, mode: string, config: TokenConfig) {
        let writeValue = JSON.stringify(config, (key, value) =>
            typeof value === 'bigint' ? value.toString() : value, 2);

        let path = rootPath + `/configs/${mode}.tokenConfig.json`;

        let fs = require('fs')
        fs.writeFileSync(path, writeValue, function (err: any) {
            if (err) {
                console.log(err);
            }
        });
    }
}